<?php

declare(strict_types=1);

namespace Drupal\Tests\partial_datelist\Functional;

use Drupal\Core\Entity\Entity\EntityFormDisplay;
use Drupal\field\Entity\FieldConfig;
use Drupal\field\Entity\FieldStorageConfig;
use Drupal\partial_datelist\PartialDatelistConfig;
use Drupal\Tests\BrowserTestBase;

/**
 * Tests the Partial Datelist module's UI integration.
 *
 * @group partial_datelist
 */
class PartialDatelistFunctionalTest extends BrowserTestBase {

  /**
   * The name of the custom content type being created.
   */
  const CONTENT_TYPE_NAME = 'partial_datelist_test';

  /**
   * The sequence of tests being executed.
   */
  const HIDDEN_ITEMS_TEST_SEQUENCE = [
    // Date only tests.
    'date' => [
      [
        'day',
      ],
      [
        'month',
        'day',
      ],
    ],

    // Date and time tests.
    'datetime' => [
      [
        'second',
        'minute',
      ],
      [
        'second',
        'minute',
        'hour',
      ],
      [
        'second',
        'minute',
        'hour',
        'day',
      ],
    ],
  ];

  const TEST_DATES = [
    [
      'year' => '1912',
      'month' => '05',
      'day' => '23',
      'hour' => '14',
      'minute' => '45',
      'second' => '30',
    ],
    [
      'year' => '1945',
      'month' => '11',
      'day' => '08',
      'hour' => '09',
      'minute' => '15',
      'second' => '00',
    ],
    [
      'year' => '1963',
      'month' => '07',
      'day' => '19',
      'hour' => '22',
      'minute' => '30',
      'second' => '45',
    ],
    [
      'year' => '1980',
      'month' => '02',
      'day' => '14',
      'hour' => '06',
      'minute' => '10',
      'second' => '20',
    ],
    [
      'year' => '1995',
      'month' => '09',
      'day' => '30',
      'hour' => '18',
      'minute' => '50',
      'second' => '10',
    ],
    [
      'year' => '1905',
      'month' => '03',
      'day' => '12',
      'hour' => '11',
      'minute' => '05',
      'second' => '55',
    ],
    [
      'year' => '1928',
      'month' => '08',
      'day' => '04',
      'hour' => '20',
      'minute' => '25',
      'second' => '40',
    ],
    [
      'year' => '2001',
      'month' => '12',
      'day' => '17',
      'hour' => '03',
      'minute' => '40',
      'second' => '05',
    ],
    [
      'year' => '2018',
      'month' => '06',
      'day' => '09',
      'hour' => '15',
      'minute' => '35',
      'second' => '25',
    ],
    [
      'year' => '2023',
      'month' => '10',
      'day' => '15',
      'hour' => '12',
      'minute' => '30',
      'second' => '00',
    ],
  ];

  /**
   * {@inheritdoc}
   */
  protected $defaultTheme = 'claro';

  /**
   * {@inheritdoc}
   */
  protected static $modules = [
    'field',
    'field_ui',
    'datetime',
    'datetime_range',
    'partial_datelist',
    'node',
  ];

  /**
   * {@inheritdoc}
   */
  protected function setUp(): void {
    parent::setUp();

    $this->drupalCreateContentType(['type' => self::CONTENT_TYPE_NAME]);

    $this->drupalLogin($this->drupalCreateUser([
      'administer content types',
      'administer node fields',
      'administer node form display',
      'administer node display',
      'administer nodes',
      'access content',
      'edit any ' . self::CONTENT_TYPE_NAME . ' content',
      'create ' . self::CONTENT_TYPE_NAME . ' content',
    ]));
  }

  /**
   * Performs multiple tests using datetime type date.
   *
   * Tests the presence of checkboxes on the widget configuration form.
   * Tests the persistence of updated widget configuration across postbacks.
   * Tests its ability to effectively hide the date parts on the node form.
   */
  public function testPartialDateWidgetAndNodeForm() {
    $this->performWidgetAndNodeTests();
  }

  /**
   * Identical to testPartialDateWidgetAndNodeForm, but tests datetime type.
   */
  public function testPartialDatetimeWidgetAndNodeForm() {
    $this->performWidgetAndNodeTests('datetime');
  }

  /**
   * Identical to testPartialDateWidgetAndNodeForm, but uses a range field.
   */
  public function testPartialDateRangeWidgetAndNodeForm() {
    $this->performWidgetAndNodeTests('date', TRUE);
  }

  /**
   * Identical to testPartialDatetimeWidgetAndNodeForm, but uses a range field.
   */
  public function testPartialDatetimeRangeWidgetAndNodeForm() {
    $this->performWidgetAndNodeTests('datetime', TRUE);
  }

  /**
   * Reusable function that performs widget and node/add tests.
   * @return void
   */
  protected function performWidgetAndNodeTests($datetime_type = 'date', $perform_range_test = FALSE):void {
    // Basic test settings.
    $field_name = "field_test_{$datetime_type}" . ($perform_range_test ? '_range' : '');
    $field_id = str_replace('_', '-', $field_name);
    $content_type_name = self::CONTENT_TYPE_NAME;
    $field_storage_type = (!$perform_range_test) ? 'datetime' : 'daterange';
    $widget_type = (!$perform_range_test) ? 'datetime_datelist' : 'daterange_datelist';
    $form_display_path = "admin/structure/types/manage/{$content_type_name}/form-display";
    $node_add_path = "node/add/{$content_type_name}";
    $datelist_parts = PartialDatelistConfig::DATELIST_PARTS;
    $datelist_parts = array_filter($datelist_parts, function ($info) use ($datetime_type) {
      return in_array($datetime_type, $info['context']);
    });
    $xpath_tr = "//tr[@id='{$field_id}']";
    $xpath_input = "//input[@name='{$field_name}_settings_edit']";
    $test_sequence = self::HIDDEN_ITEMS_TEST_SEQUENCE[$datetime_type];
    $datelist_settings_class_name = PartialDatelistConfig::DATELIST_SETTINGS_CLASS_NAME;

    // Create a field storage with settings to validate.
    $field_storage = FieldStorageConfig::create([
      'field_name' => $field_name,
      'entity_type' => 'node',
      'type' => $field_storage_type,
      'settings' => ['datetime_type' => $datetime_type],
    ]);
    $field_storage->save();

    $field = FieldConfig::create([
      'field_storage' => $field_storage,
      'bundle' => self::CONTENT_TYPE_NAME,
    ]);
    $field->save();

    // Load or create the form display config for the content type.
    $form_display = EntityFormDisplay::load('node.' . self::CONTENT_TYPE_NAME . '.default');
    if (!$form_display) {
      $form_display = EntityFormDisplay::create([
        'targetEntityType' => 'node',
        'bundle' => self::CONTENT_TYPE_NAME,
        'mode' => 'default',
        'status' => TRUE,
      ]);
    }

    // Set the widget and move it into the visible region.
    $form_display->setComponent($field_name, [
      'type' => $widget_type,
      'weight' => -50,
      'region' => 'content',
    ]);
    $form_display->save();

    // Update the field configuration according to the test sequence.
    foreach ($test_sequence as $test_sequence_item) {
      // Navigate to the form display settings page.
      $this->drupalGet($form_display_path);

      $page = $this->getSession()->getPage();

      // Check if the field row exists in the form display table.
      $tr_element = $page->find('xpath', $xpath_tr);
      $this->assertNotEmpty(
        $tr_element,
        "The field '{$field_name}' was not found in the form display table. This is unexpected behavior.",
      );

      // Check if the input button that opens the widget details configuration is present.
      $input_element = $tr_element->find('xpath', $xpath_input);
      $this->assertNotEmpty(
        $input_element,
        "The input element associated with field '{$field_name}' was not found within the form display table. This is unexpected behavior.",
      );

      // Open the date field widget configuration.
      $input_element->click();

      // Ensure the partial datelist checkboxes control exists on the page.
      $hidden_datelist_parts = $tr_element->find('xpath', "//fieldset[contains(@class, '{$datelist_settings_class_name}')]");
      if (!$hidden_datelist_parts) {
        $this->fail("The datelist settings checkboxes could not be found.");
      }

      // Stores the found checkboxes within form display.
      $checkboxes = [];

      // Assert each of the date parts exist.
      foreach ($datelist_parts as $key_part => $datelist_part) {
        $checkbox = $hidden_datelist_parts->find('xpath', "//input[@value='{$datelist_part['setting']}']");

        // Assert and exit loop if a checkbox is not found.
        $this->assertNotEmpty($checkbox, "The checkbox for '{$datelist_part['label']}' was not found, this is unexpected behavior.");
        if (!$checkbox) {
          continue;
        }

        // Append a new value to the checkboxes array.
        $checkboxes["fields[{$field_name}][settings_edit_form][third_party_settings][partial_datelist][hidden_datelist_parts][{$datelist_part['setting']}]"] = in_array($key_part, $test_sequence_item);
      }

      // Submit the form to save the changes.
      $this->submitForm($checkboxes, 'Save');

      // Request the form display settings page again to check the changes.
      $this->drupalGet($form_display_path);
      $callback_page = $this->getSession()->getPage();
      $callback_tr_element = $callback_page->find('xpath', $xpath_tr);
      $callback_input_element = $callback_tr_element->find('xpath', $xpath_input);
      $callback_input_element->click();

      // Assert items are checked correctly according to the test sequence.
      foreach ($datelist_parts as $key_part => $datelist_part) {
        $checkbox = $hidden_datelist_parts->find('xpath', "//input[@value='{$datelist_part['setting']}']");

        // Exit the loop if a checkbox is not found.
        if (!$checkbox) {
          continue;
        }

        // Revisit the field widget setting and test if the checkbox is checked
        // according to the test sequence plan.
        $is_checked = $checkbox->getAttribute('checked') === 'checked';
        $should_be_checked = in_array($key_part, $test_sequence_item);
        $is_checked_string = var_export($is_checked, TRUE);
        $should_be_checked_string = var_export($should_be_checked, TRUE);
        $this->assertTrue($is_checked === $should_be_checked, "There is a mismatch between the checkbox value for field '{$datelist_part['label']}' and the value that should be assigned per the test sequence. It should be set as '{$should_be_checked_string}', but it was found to be '{$is_checked_string}' instead.");
      }

      $activated_datelist_parts = array_filter($datelist_parts, function ($key) use ($test_sequence_item) {
        return !in_array($key, $test_sequence_item);
      }, ARRAY_FILTER_USE_KEY);

      // Test the node add form.
      $this->performNodeFormTests(
        $node_add_path,
        $activated_datelist_parts,
        $field_name,
        $perform_range_test,
        TRUE
      );

      // Extract the node ID from the current URL.
      $current_url = $this->getSession()->getCurrentUrl();
      preg_match('/node\/(\d+)/', $current_url, $matches);
      $node_id = $matches[1] ?? NULL;

      // Assert that the node ID was retrieved successfully.
      $this->assertNotEmpty($node_id, 'The node ID could not be retrieved from the URL after having saved a test node.');

      // Assert that the edit form is loaded.
      $this->assertSession()->statusCodeEquals(200);

      // Test the node edit form.
      $this->performNodeFormTests(
        "node/{$node_id}/edit",
        $activated_datelist_parts,
        $field_name,
        $perform_range_test
      );
    }
  }

  /**
   * Reusable function that performs tests on add/edit node forms.
   *
   * @param string $node_path
   *   The path of the node page being tested (add or edit).
   * @param array $activated_datelist_parts
   *   The test sequence item to check against.
   * @param string $field_name
   *   The name of the field being tested.
   * @param bool $perform_range_test
   *   Whether to perform a range test.
   * @param bool $perform_save
   *   Whether to perform a node save operation.
   *
   * @return void
   */
  protected function performNodeFormTests(string $node_path, array $activated_datelist_parts, string $field_name, bool $perform_range_test, bool $perform_save = FALSE):void {

    // Retrieves a random date from a set of predefined test dates.
    $test_date = self::TEST_DATES[rand(0, count(self::TEST_DATES) - 1)];

    // Set the content edit array with a dummy title.
    $content_edit = [
      'title[0][value]' => 'Test Node Title',
    ];

    // Navigate to the node path for add/edit.
    $this->drupalGet($node_path);
    $node_page = $this->getSession()->getPage();

    // Find all the select elements for the date parts.
    $selects = $node_page->findAll('xpath', "//select[starts-with(@name, '{$field_name}[0][value]')]");
    $this->assertEquals(count($activated_datelist_parts), count($selects), 'The number of select elements on the node/add page does not match the number of activated datelist parts for the current test sequence.');

    // Assert items are checked correctly according to the test sequence.
    foreach ($activated_datelist_parts as $key_part => $datelist_part) {
      $select = $node_page->find('xpath', "//select[@name='{$field_name}[0][value][{$key_part}]']");
      $this->assertNotEmpty($select, "The select element for '{$datelist_part['label']}' was not found in the node add form, this is unexpected behavior.");

      // If a range test is activated, check for the end date select element.
      if ($perform_range_test) {
        $select = $node_page->find('xpath', "//select[@name='{$field_name}[0][end_value][{$key_part}]']");
        $this->assertNotEmpty($select, "The daterange select element for '{$datelist_part['label']}' associated with capturing an end date was not found in the node add form, this is unexpected behavior.");
      }

      // Update the field values.
      if ($perform_save) {
        $content_edit["{$field_name}[0][value][{$key_part}]"] = $test_date[$key_part];

        if ($perform_range_test) {
          $content_edit["{$field_name}[0][end_value][{$key_part}]"] = $test_date[$key_part];
        }
      }
    }

    // Submit the form.
    if ($perform_save) {
      $this->submitForm($content_edit, 'Save');
    }
  }

}
